/*
 * can.c
 *
 *  Created on: 17.03.2012
 *      Author: Frank
 */

#include "can.h"


void mcp2515_write_register( uint8_t adress, uint8_t data )
{
    // /CS des MCP2515 auf Low ziehen
    PORTB &= ~(1 << SS);

    spi_putc(SPI_WRITE);
    spi_putc(adress);
    spi_putc(data);

    // /CS Leitung wieder freigeben
    PORTB |= (1 << SS);
}


uint8_t mcp2515_read_register(uint8_t adress)
{
    uint8_t data;

    // /CS des MCP2515 auf Low ziehen
    PORTB &= ~(1 << SS);

    spi_putc(SPI_READ);
    spi_putc(adress);

    data = spi_putc(0xFF);

    // /CS Leitung wieder freigeben
    PORTB |= (1 << SS);

    return data;
}


void mcp2515_bit_modify(uint8_t adress, uint8_t mask, uint8_t data)
{
    // /CS des MCP2515 auf Low ziehen
    PORTB &= ~(1 << SS);

    spi_putc(SPI_BIT_MODIFY);
    spi_putc(adress);
    spi_putc(mask);
    spi_putc(data);

    // /CS Leitung wieder freigeben
    PORTB |= (1 << SS);
}


void init_can()
{
	//Controller resetten
    PORTB &= ~(1 << SS);
    spi_putc(SPI_RESET);
    _delay_ms(1);
    PORTB |= (1 << SS);

    _delay_ms(10);
	// CAN Bitrate 125 kbps

	mcp2515_write_register(CNF1, (1 << BRP0) | (1 << BRP1) | (1 << BRP2)); //ok
	mcp2515_write_register(CNF2, (1 << BTLMODE) | (1 << PHSEG11));//ok
	mcp2515_write_register(CNF3, (1 << PHSEG21));//ok

	// Aktivieren der Rx Buffer Interrupts
	mcp2515_write_register(CANINTE, (1 << RX1IE) | (1 << RX0IE));//ok

	// Buffer 0 : Empfangen aller Nachrichten
	mcp2515_write_register(RXB0CTRL, (1 << RXM1) | (1 << RXM0)); //ok

	// Buffer 1 : Empfangen aller Nachrichten
	mcp2515_write_register(RXB1CTRL, (1 << RXM1) | (1 << RXM0));
	//ok ->RSM1 Schreibfehler in datasheet

    // Alle Bits der Empfangsmaske loeschen,
    // damit werden alle Nachrichten empfangen

    mcp2515_write_register(RXM0SIDH, 0);
    mcp2515_write_register(RXM0SIDL, 0);
    mcp2515_write_register(RXM0EID8, 0);
    mcp2515_write_register(RXM0EID0, 0);

    mcp2515_write_register(RXM1SIDH, 0);
    mcp2515_write_register(RXM1SIDL, 0);
    mcp2515_write_register(RXM1EID8, 0);
    mcp2515_write_register(RXM1EID0, 0);
    //alle ok

    /*
     *  Einstellen der Pin Funktionen
     */

    // Deaktivieren der Pins RXnBF Pins (High Impedance State)
    mcp2515_write_register( BFPCTRL, 0 );

    // TXnRTS Bits als Inputs schalten
    mcp2515_write_register( TXRTSCTRL, 0 );

    // Device zurueck in den normalen Modus versetzten
    mcp2515_bit_modify( CANCTRL, 0xE0, 0);
    //alle ok
}


void can_send_message(CANMessage *p_message)
{
    uint8_t length = p_message->length;

    // ID einstellen
    mcp2515_write_register(TXB0SIDH, (uint8_t) (p_message->id>>3));
    mcp2515_write_register(TXB0SIDL, (uint8_t) (p_message->id<<5));

    // Ist die Nachricht ein "Remote Transmit Request"
    if (p_message->rtr)
    {
        /* Eine RTR Nachricht hat zwar eine Laenge,
           aber keine Daten */

        // Nachrichten Laenge + RTR einstellen
        mcp2515_write_register(TXB0DLC, (1<<RTR) | length);
    }
    else
    {
        // Nachrichten Laenge einstellen
        mcp2515_write_register(TXB0DLC, length);

        // Daten
        for (uint8_t i = 0; i < length; i++) {
            mcp2515_write_register(TXB0D0 + i, p_message->data[i]);
        }
    }

    // CAN Nachricht verschicken
    PORTB &= ~(1 << SS);
    spi_putc(SPI_RTS | 0x01);
    PORTB |= (1 << SS);
}


static uint8_t mcp2515_read_rx_status(void)
{
    uint8_t data;

    // /CS des MCP2515 auf Low ziehen
    PORTB &= ~(1 << SS);

    spi_putc(SPI_RX_STATUS);
    data = spi_putc(0xff);

    // Die Daten werden noch einmal wiederholt gesendet,
    // man braucht also nur eins der beiden Bytes auswerten.
    spi_putc(0xff);

    // /CS Leitung wieder freigeben
    PORTB |= (1 << SS);

    return data;
}


uint8_t can_get_message(CANMessage *p_message)
{
    // Status auslesen
    uint8_t status = mcp2515_read_rx_status();

    if (status & (1 << 6))
    {
        // Nachricht in Puffer 0

        PORTB &= ~(1 << SS);    // CS Low
        spi_putc(SPI_READ_RX);
    }
    else if (status & (1 << 7))
    {
        // Nachricht in Puffer 1

        PORTB &= ~(1 << SS);    // CS Low
        spi_putc(SPI_READ_RX | 0x04);
    }
    else {
        /* Fehler: Keine neue Nachricht vorhanden */
        return 0xff;
    }

    // Standard ID auslesen
    p_message->id =  (uint16_t) spi_putc(0xff) << 3;
    p_message->id |= (uint16_t) spi_putc(0xff) >> 5;

    spi_putc(0xff);
    spi_putc(0xff);

    // Laenge auslesen
    uint8_t length = spi_putc(0xff) & 0x0f;
    p_message->length = length;

    // Daten auslesen
    for (uint8_t i = 0; i < length; i++)
    {
        p_message->data[i] = spi_putc(0xff);
    }

    PORTB |= (1 << SS);

    if (status & (1 << 3))
    {
        p_message->rtr = 1;
    } else
    {
        p_message->rtr = 0;
    }

    // Interrupt Flag loeschen
    if (status & (1 << 6))
    {
        mcp2515_bit_modify(CANINTF, (1 << RX0IF), 0);
    }
    else
    {
        mcp2515_bit_modify(CANINTF, (1 << RX1IF), 0);
    }

    return (status & 0x07);
}
